#define YGGMAXTOKENLEN 32
#define YGGCHECKSUM		'Y'

typedef struct YggNode {
	unsigned int checksum 	:8;
	unsigned int contents 	:4;
	unsigned int parenttype :4;
	ClientData data;
	char name[YGGMAXTOKENLEN];
	struct YggNode *parent;
	struct YggNode *itree;
	struct YggNode *jtree;
	struct YggNode *nextPtr;
} YggNode;

typedef struct YggEnv {
	Tcl_Interp	*interp;
	struct YggNode 	*root;
	struct YggNode 	*trash;
	Tcl_Namespace	*globalNsPtr;
	int		nodecount;
	struct YggEnv	*nextPtr;
} YggEnv;

/*
 * Stored a linear path through a tree
 */
typedef struct YggPath {
	char *name;
	int relation;
	struct YggPath *nextPtr;
} YggPath;

/*
 * Cached Set of resolved variables
 */
typedef struct YggVarTable {
	YggNode *node;
	Tcl_Var		*varinfo;
	struct YggVarTable *nextPtr;
} YggVarTable;

/*
 * Macros
 */

#define Ygg_NewPath(ptr,nameval,relationval) { \
	ptr = (YggPath *)ckalloc((sizeof(YggPath))); \
	ptr->name = nameval; \
	ptr->relation = relationval; \
	ptr->nextPtr = NULL; \
}

#define Insert(headptr,thisptr) { \
	thisptr->nextPtr = headptr; \
	headptr = thisptr; \
}

#define Swap(temp,a) { \
	temp = a; 			\
	a = a->nextPtr; 	\
	a->nextPtr = temp;	\
}
/*
 * Declarations
 */

/* Exported Functions */
extern int Ygg_Init _ANSI_ARGS_((Tcl_Interp *interp));
extern int Ygg_ObjCmd _ANSI_ARGS_ ((
	ClientData dummy, Tcl_Interp *interp, 
	int tobjc, Tcl_Obj *CONST tobjv[]));

extern int Ygg_ObjExtraCmd _ANSI_ARGS_ ((
	ClientData dummy, Tcl_Interp *interp, 
	int tobjc, Tcl_Obj *CONST tobjv[]));
	
extern YggNode * 
	Ygg_Record _ANSI_ARGS_((char *name,int namelen,int create));
extern char * 
	Ygg_Address _ANSI_ARGS_((YggNode *record));

extern void Ygg_NodeFreeData _ANSI_ARGS_ ((YggNode *node));

extern int Ygg_TreeLink _ANSI_ARGS_ ((
	Tcl_Interp *interp, char *name, char *ptr, int code));
	
extern int 
	Ygg_Operation _ANSI_ARGS_((
		char *src,int srclen,
		char *dest,int destlen,
		int opcode));

extern void Ygg_SetObj _ANSI_ARGS_((YggNode *node,Tcl_Obj *objptr));
extern Tcl_Obj * Ygg_GetObj _ANSI_ARGS_((YggNode *node));

extern Tcl_DString * 
	Ygg_Export _ANSI_ARGS_((YggNode *node,int root,int recurse));

extern int 
	Ygg_Import _ANSI_ARGS_((YggNode *node,char *buffer));
	
extern void Ygg_Error _ANSI_ARGS_((int errCode,char *info));
/*
 * Enumerated Node Contents
 */
enum {
	yggEmpty =0,
	yggTclObj,
	yggLink
};

/*
 * Enumerated Link Types
 * Used for storage and search
 */
enum {
	yggError = -1,
	yggContinue = 0,
	yggItree=1,
	yggJtree=2,
	yggKtree,		
	yggHead,
	yggEnd,
};

enum {
	/*
	 * Enumerated Structure Opcodes
	 */
	yggOpCopy=1,
	yggOpLink=2,
	yggOpMove=3,
	/* 
	 * Enumerated Options
	 */
	yggOpForce=16
};
